#ifdef _UNICODE
#define UNICODE
#undef _MBCS
#undef MBCS
#endif

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <tchar.h>

#ifdef _UNICODE
#define APP_NAME L"ml-ent v0.3.1 (Unicode build)"
#else
#define APP_NAME "ml-ent v0.3.1 (Non-Unicode build)"
#endif
/*
ver0.3
[NEW] '/' to "&#47;"
[NEW] Input text length unlimted
[NEW] Uses Unicode whenever possible
[NEW] [ScrollLock] to escape non-ascii chars as &#x....;

ver 0.3.1 2006-12-05
[NEW] Silent/Verbose
[NEW] Enable/Disable the "// -> /&#47;" conv
*/


// How big the output buffer should be?
// 16 means, we prepare 16 * LEN size output buffer for input length LEN
// Output is generally longer than input, because of 'escaping'
// The 'worst' case is 10 or 11, because:
// &#x10FFFF;
// &#1114111;
// &#xFFFFFF;  (10-letter)
// &#16777215; (11-letter)
// So, 16 is big enough.

#define SUNNY_LUCHE 16

#define MAX_TAB SUNNY_LUCHE
// Which means, input TAB gets upto 16 SPACEs (default = 4 SPACEs)
// This doesn't cause the output buffer overflow, because of the way we defined SUNNY_LUCHE 



inline void Info( LPCTSTR msg );
inline BOOL IsScrollLocked( void );
inline BOOL IsScrollLockDown( void );
void HtmlEntitiesA( char * lpszInput, char * lpszOutput );
int InitializeConvParams(
	LPSTR lpszQuot, LPSTR lpszApos, LPSTR lpszBr, LPSTR lpszTab,
	BOOL * pbDoubleSlash, BOOL * pbSilent);
int ParseLine( char *buf, char *szVariable );


#ifdef _UNICODE
#define SURROGATE_HI( wc ) ( 0xD800 <= (wc) && (wc) <= 0xDBFF )
#define SURROGATE_LO( wc ) ( 0xDC00 <= (wc) && (wc) <= 0xDFFF )
#define SURROGATE_OFFSET 0xFCA02400UL
// = 0x10000 - (0xD800 << 10) - 0xDC00
inline int ParseSurrogate( const WCHAR * p );
void HtmlEntitiesW( WCHAR * lpszInput, WCHAR * lpszOutput );
inline BOOL IsAnsi( LPCWSTR strW );
#endif // ifdef _UNICODE



BOOL bSilent = FALSE;


int  WINAPI WinMain( HINSTANCE hInstance,
					 HINSTANCE Fine, LPSTR Rein, int Poomo )
{
	UNREFERENCED_PARAMETER( hInstance );
	UNREFERENCED_PARAMETER( Fine );
	UNREFERENCED_PARAMETER( Rein );
	UNREFERENCED_PARAMETER( Poomo );


#ifdef _UNICODE		
	UINT PriorityList[ 2 ] = { CF_UNICODETEXT, CF_TEXT };
	const int iClipFormat = GetPriorityClipboardFormat( PriorityList, 2 );
	if( iClipFormat != CF_UNICODETEXT && iClipFormat != CF_TEXT )
	{
		Info(_T("No text found in Clipboard."));
		return 1;
	}
#else
	const int iClipFormat = CF_TEXT;
	if( ! IsClipboardFormatAvailable( iClipFormat ) )	
	{
		Info(_T("No text found in Clipboard."));
		return 1;
	}
#endif

	
	if( ! OpenClipboard(NULL) ) return 1;


#ifdef _UNICODE
	if( iClipFormat == CF_UNICODETEXT )
	{
		HANDLE hClipW = GetClipboardData( CF_UNICODETEXT );
		if( hClipW == NULL )
		{
			CloseClipboard();
			return 1;
		}

		WCHAR * lpszClipTextW = ( WCHAR * ) GlobalLock( hClipW );
		if( lpszClipTextW == NULL )
		{
			GlobalUnlock( hClipW );
			CloseClipboard();
			return 1;
		}

		const int iInputlen = lstrlenW( lpszClipTextW );

		if( iInputlen <= 0 )
		{
			GlobalUnlock( hClipW );
			CloseClipboard();
			return 1;
		}
		
		WCHAR * lpszInputW = new WCHAR[ iInputlen + 1 ];
		lstrcpyW( lpszInputW, lpszClipTextW );
		GlobalUnlock( hClipW );

		EmptyClipboard();

		// SUNNY_LUCHE * sizeof(WCHAR) for each input WCHAR should be roughly enough,
		// but technically we need +2 for the terminating L'\0'.
		// Let's add 1024 instead of 2, so that we don't need to worry about minor details.
		HGLOBAL hNewText = GlobalAlloc(
			GHND, iInputlen * SUNNY_LUCHE * sizeof(WCHAR) + 1024 );

		if( hNewText == NULL )
		{
        	delete [] lpszInputW;
			return 1;
		}

		WCHAR * lpszOutputW = ( WCHAR * ) GlobalLock( hNewText );
		HtmlEntitiesW( lpszInputW, lpszOutputW );

		delete [] lpszInputW;
		
		const int iOutputlen = lstrlenW( lpszOutputW );

		const int iBufSizeNeeded = WideCharToMultiByte( CP_ACP, WC_NO_BEST_FIT_CHARS,
			lpszOutputW,
			-1,
			NULL,
			0,
			NULL,
			NULL
		);

		HGLOBAL hMem = GlobalAlloc( GHND, iBufSizeNeeded + 1 );
		char * pChar = ( char * ) GlobalLock( hMem );
		BOOL bFailed = FALSE;
		WideCharToMultiByte( CP_ACP, WC_NO_BEST_FIT_CHARS,
			lpszOutputW,
			-1,
			pChar,
			iBufSizeNeeded,
			NULL,
			&bFailed
		);
		GlobalUnlock( hMem );
		SetClipboardData( CF_TEXT, hMem );

		GlobalUnlock( hNewText );

		SetClipboardData( CF_UNICODETEXT, hNewText );
		CloseClipboard();

		WCHAR szResult[ 1024 ];
        wsprintf( szResult,
        	L"Iuput length: %d\r\n"
			L"Output length: %d\r\n"
			L"%s\r\n"
			, iInputlen
			, iOutputlen
			, bFailed
				? L"CF_UNICODETEXT Only"
				: L"Success!"
		);
		if( ! bSilent )
		MessageBox( NULL, szResult, APP_NAME,
			MB_OK | ( bFailed? MB_ICONEXCLAMATION : MB_ICONINFORMATION )
		);
	}
	else
#endif		
	// iClipFormat == CF_TEXT
	{
		HANDLE hClipA = GetClipboardData( CF_TEXT );
		if( hClipA == NULL )
		{
			CloseClipboard();
			return 1;
		}

		char * lpszClipTextA = ( char * ) GlobalLock( hClipA );
		if( lpszClipTextA == NULL )
		{
			GlobalUnlock( hClipA );
			CloseClipboard();
			return 1;
		}

		const int iInputlen = lstrlenA( lpszClipTextA );

		if( iInputlen <= 0 )
		{
			GlobalUnlock( hClipA );
			CloseClipboard();
			return 1;
		}

		char * lpszInputA = new char[ iInputlen + 1 ];
		lstrcpyA( lpszInputA, lpszClipTextA );

		GlobalUnlock( hClipA );

		EmptyClipboard();

		HGLOBAL hNewText = GlobalAlloc( 
			GHND, iInputlen * SUNNY_LUCHE * sizeof( char ) + 1024 );
		
		if( hNewText == NULL )
		{
        	delete [] lpszInputA;
			return 1;
		}

		char * lpszOutputA = ( char * ) GlobalLock( hNewText );
		HtmlEntitiesA( lpszInputA, lpszOutputA );
		const int iOutputlen = lstrlenA( lpszOutputA );
		GlobalUnlock( hNewText );
		delete [] lpszInputA;

		SetClipboardData(CF_TEXT, hNewText );
		CloseClipboard();

		TCHAR szResult[ 1024 ];
        wsprintf( szResult,
        	_T("Iuput length: %d\r\nOutput length: %d"),
			iInputlen, iOutputlen);
		if( ! bSilent ) Info( szResult );
    }

	return 1;
}

void HtmlEntitiesA( char * lpszInput, char * lpszOutput )
{
	if( lpszInput == NULL || lpszOutput == NULL ) return;

    lstrcpyA( lpszOutput, "" );

    if( lstrcmpA( lpszInput, "v" ) == 0 )
	{
    	char szVersionInfo[ 1024 ];
        sprintf( szVersionInfo,
			"%s\r\nhttp://www.faireal.net/mion/\r\nCompiled on %s (%s UTC)",
        	APP_NAME, __DATE__, __TIME__
		);
#ifndef _UNICODE
        Info( szVersionInfo );
#endif
        lstrcpyA( lpszOutput, szVersionInfo );
        return;
    }

	// Default Values to be used when .ini file in unavailable...
	char lpszQuot[ SUNNY_LUCHE ] = "&quot;"; // double quotation to what?
	char lpszApos[ SUNNY_LUCHE ] = "&#39;"; // single quotation to what?
	char lpszBr[ SUNNY_LUCHE ] = "\r\n" ; // new line to <br> ?
	char lpszTab[ SUNNY_LUCHE ] = "    "; // 4 spaces

	char lpszDoubleHyphen[ SUNNY_LUCHE ] = "-&#45;";

	BOOL bDoubleSlash = FALSE;
	InitializeConvParams( lpszQuot, lpszApos, lpszBr, lpszTab, 
		&bDoubleSlash, &bSilent );

	int iInputlen = lstrlenA( lpszInput );

	for( int i=0; i < iInputlen; i++ )
	{
    	if(lpszInput[i] == '&') lstrcatA(lpszOutput, "&amp;");
    	else if(lpszInput[i] == '<') lstrcatA(lpszOutput, "&lt;");
        else if(lpszInput[i] == '>') lstrcatA(lpszOutput, "&gt;");
        else if(lpszInput[i] == '"') lstrcatA(lpszOutput, lpszQuot );
        else if(lpszInput[i] == '\'') lstrcatA(lpszOutput, lpszApos );
        else if(lpszInput[i] == '\r' && lpszInput[i+1] == '\n' )
		{
			lstrcatA(lpszOutput, lpszBr);
			i++;
		}
		else if( bDoubleSlash && lpszInput[i] == '/' && lpszInput[i+1] == '/' )
		{
			lstrcatA(lpszOutput, "/&#47;" );
			i++;
		}
        else if(lpszInput[i] == '/'	&& lpszInput[i+1] == '*' )
		{
			lstrcatA(lpszOutput, "&#47;*" );
			i++;
		}
        else if(lpszInput[i] == '*'	&& lpszInput[i+1] == '/' )
		{
			lstrcatA(lpszOutput, "*&#47;" );
			i++;
		}
        else if(lpszInput[i] == '\n' || lpszInput[i] == '\r' ) lstrcatA(lpszOutput, lpszBr);
        else if(lpszInput[i] == '\t') lstrcatA(lpszOutput, lpszTab );
		else if(lpszInput[i] == '-' && lpszInput[i+1] == '-' )
		{
			lstrcatA( lpszOutput, lpszDoubleHyphen );
			i++;
		}
        else
		{
            const int last = lstrlenA(lpszOutput)-1;
    		lpszOutput[ last+1 ] = lpszInput[ i ];
            lpszOutput[ last+2 ] = '\0';
		}
	}
}


#ifdef _UNICODE
void HtmlEntitiesW( WCHAR * lpszInput, WCHAR * lpszOutput )
{
	if( lpszInput == NULL || lpszOutput == NULL ) return;


	const BOOL bAscii = IsScrollLockDown();
	const BOOL bAnsi  = ! bAscii && IsScrollLocked();

    lstrcpyW( lpszOutput, L"" );

    if( lstrcmpW( lpszInput, L"v" ) == 0 )
	{
    	WCHAR szVersionInfo[ 1024 ];
        wsprintf( szVersionInfo,
			_T("%s\r\nhttp://www.faireal.net/mion/\r\nCompiled on %hs (%hs UTC)"),
        	APP_NAME, __DATE__, __TIME__
		);
        Info( szVersionInfo );
        lstrcpyW( lpszOutput, szVersionInfo );
        return;
    }

	// Default Values to be used when .ini file in unavailable...
	char lpszQuotA[ SUNNY_LUCHE ] = "&quot;"; // double quotation to what?
	char lpszAposA[ SUNNY_LUCHE ] = "&#39;"; // single quotation to what?
	char lpszBrA[ SUNNY_LUCHE ] = "\r\n" ; // new line to <br> ?
	char lpszTabA[ SUNNY_LUCHE ] = "    "; // 4 spaces
	char lpszDoubleHyphenA[ SUNNY_LUCHE ] = "-&#45;";

	BOOL bDoubleSlash = FALSE;
	
	InitializeConvParams( lpszQuotA, lpszAposA, lpszBrA, lpszTabA, 
		&bDoubleSlash, &bSilent );

			WCHAR dbg[1000];
			wsprintf(dbg,L"bDoubleSlash=%d\r\n",bDoubleSlash);
			OutputDebugString(dbg);


#define AtoW( strA, strW ) wsprintfW( strW, L"%hs", strA )

	WCHAR lpszQuot[ SUNNY_LUCHE ];
	AtoW( lpszQuotA, lpszQuot );
	WCHAR lpszApos[ SUNNY_LUCHE ];
	AtoW( lpszAposA, lpszApos );
	WCHAR lpszBr[ SUNNY_LUCHE ];
	AtoW( lpszBrA, lpszBr );
	WCHAR lpszTab[ SUNNY_LUCHE ];
	AtoW( lpszTabA, lpszTab );
	WCHAR lpszDoubleHyphen[ SUNNY_LUCHE ];
	AtoW( lpszDoubleHyphenA, lpszDoubleHyphen );

	int iInputlen = lstrlenW( lpszInput );


	register int i;
	for( i=0; i < iInputlen; i++ )
	{
    	if(lpszInput[i] == L'&') lstrcatW(lpszOutput, L"&amp;");
    	else if(lpszInput[i] == L'<') lstrcatW(lpszOutput, L"&lt;");
        else if(lpszInput[i] == L'>') lstrcatW(lpszOutput, L"&gt;");
        else if(lpszInput[i] == L'"') lstrcatW(lpszOutput, lpszQuot );
        else if(lpszInput[i] == L'\'') lstrcatW(lpszOutput, lpszApos );
        else if(lpszInput[i] == L'\r' && lpszInput[i+1] == L'\n' )
		{
			lstrcatW(lpszOutput, lpszBr);
			i++;
		}
        else if( bDoubleSlash && lpszInput[i] == L'/' && lpszInput[i+1] == L'/' )
		{
			lstrcatW(lpszOutput, L"/&#47;" );
			i++;
		}
        else if(lpszInput[i] == L'/' && lpszInput[i+1] == L'*' )
		{
			lstrcatW(lpszOutput, L"&#47;*" );
			i++;
		}
        else if(lpszInput[i] == L'*' && lpszInput[i+1] == L'/' )
		{
			lstrcatW(lpszOutput, L"*&#47;" );
			i++;
		}
        else if(lpszInput[i] == L'\n' || lpszInput[i] == L'\r' ) lstrcatW(lpszOutput, lpszBr);
        else if(lpszInput[i] == L'\t') lstrcatW(lpszOutput, lpszTab );
		else if(lpszInput[i] == L'-' && lpszInput[i+1] == L'-' )
		{
			lstrcatW( lpszOutput, lpszDoubleHyphen );
			i++;
		}
		else if( bAnsi && ! IsAnsi( &lpszInput[ i ] ) || bAscii && lpszInput[i] >= 0x0080 )
		{
			WCHAR tmpstr[ SUNNY_LUCHE * 2 ]; // * 2 for possible Surrogate

			const int iUTF32 = ParseSurrogate( &lpszInput[ i ] );
			if( iUTF32 < 0 ) break; // Error: ill-formed UTF-16 detected
			else if( iUTF32 > 0 ) // Surrogate detected
			{
				wsprintf( tmpstr, L"&#x%06x;", (UINT) iUTF32 );
				i++;
			}
			else
			{
				wsprintf( tmpstr, L"&#x%04x;", (UINT) lpszInput[i] );
			}

			lstrcat( lpszOutput, tmpstr );
		}
        else
		{
            const int last = lstrlenW(lpszOutput)-1;
    		lpszOutput[ last+1 ] = lpszInput[ i ];
			// Actually we don't need to do this,
			// as we set GHND (=GMEM_MOVEABLE and *GMEM_ZEROINIT*)
			// But just in case.
            lpszOutput[ last+2 ] = L'\0';
		}
	}
}

inline BOOL IsAnsi( LPCWSTR strW )
{
	char strA[ 16 ];

	BOOL bFailed = FALSE;

	WideCharToMultiByte( CP_ACP, WC_NO_BEST_FIT_CHARS,
		strW,
		1, // check this one WCHAR only
		strA,
		15,
		NULL,
		&bFailed
	);

	return ! bFailed;
}


inline int ParseSurrogate( const WCHAR * p )
{
	if( SURROGATE_LO( *p ) || SURROGATE_HI( *p ) && ! SURROGATE_LO( *(p+1) ) )
	{
		Info( L"ERROR: An unpaired surrogate is encountered." );
		return -1;
	}
	else if( SURROGATE_HI( *p ) && SURROGATE_LO( *(p+1) ) )
	{
		return (int)(   (((DWORD) *p) << 10) + *(p+1) + SURROGATE_OFFSET );
	}
	else return 0;
}


#endif // ifdef _UNICODE




int InitializeConvParams(
	LPSTR lpszQuot, LPSTR lpszApos, LPSTR lpszBr, LPSTR lpszTab, 
	BOOL * pbDoubleSlash,
	BOOL * pbSilent

)
{
	FILE *fp=fopen("ml-ent.ini", "r");
    if(!fp) return 0;
#define BUFSIZE 1024
    char buf[ BUFSIZE ];
	int i;
    for( i=0; i< 200 ; i++) {
    	if( fgets( buf, BUFSIZE, fp) == NULL ) {
			break;
        }

		if( buf[0] == '#' || buf[0] == '\r' || buf[0] == '\n' ) continue;

		char szVariable[32] = "";
        int iValue = ParseLine( buf, szVariable );

        if( strcmp( szVariable , "quot" ) == 0 ) {
        	iValue == 0 ?
	        	lstrcpyA( lpszQuot, "\"" ) : lstrcpyA( lpszQuot, "&quot;" ) ;
		} else if( strcmp( szVariable , "apos" ) == 0 ) {
        	iValue == 0 ?
	        	lstrcpyA( lpszApos, "'" ) :
            iValue == 1 ?
                lstrcpyA( lpszApos, "&apos;" ) : lstrcpyA( lpszApos, "&#39;" ) ;
        } else if( strcmp( szVariable, "nl2br") == 0 ) {
			iValue == 0 ?
            	lstrcpyA( lpszBr , "\r\n" ) :
            iValue == 1 ?
            	lstrcpyA( lpszBr , "<br>\r\n" ) :
            iValue == 2	?
				lstrcpyA( lpszBr, "<BR>\r\n" ) : lstrcpyA( lpszBr, "<br />\r\n" );
        } else if( strcmp( szVariable , "tab" ) == 0 ) {
            if( iValue > 0 && iValue <= MAX_TAB ) {
				int j;
				for( j=0; j<iValue; j++ ) lpszTab[j] = ' '; // tab 2 spaces
				lpszTab[ iValue ] = '\0';
			} else lstrcpyA( lpszTab, "\t" );
        }
		else if( strcmpi( szVariable, "dslash" ) == 0 )
		{
			*pbDoubleSlash = !!( iValue );
		}
		else if( strcmpi( szVariable, "silent" ) == 0 )
		{
			*pbSilent = !!( iValue );
		}
    }

    fclose(fp);
  	return 1;
}

ParseLine( char *buf, char *szVariable )
{
    char *p1 = szVariable;
	char szValue[32];
    char *p2 = szValue;

    buf[32] = '\0'; // see first 32 bytes only

	int i, iValueFlag = 0;
    for( i=0; i<30; i++ ) {
		if( buf[i]=='\0' ) break;
        else if( buf[i] == ' ' && iValueFlag == 0 ) continue;
        else if( buf[i] == '\t' && iValueFlag == 0 ) continue;
		else if( buf[i] == '=' ) iValueFlag = 1;
		else if( iValueFlag == 0 ) *p1++ = buf[i];
		else *p2++ = buf[i];
	}
    *p1 = '\0', *p2 = '\0';

	return atoi( szValue );
}

inline void Info( LPCTSTR msg )
{
	MessageBox( NULL, msg, APP_NAME, MB_OK|MB_ICONINFORMATION );
}

inline BOOL IsScrollLocked( void )
{
	return ((WORD) GetKeyState( VK_SCROLL )) & 0x0001;
}

inline BOOL IsScrollLockDown( void )
{
	return !! ( ((WORD) GetKeyState( VK_SCROLL )) & 0x8000 );
}

